import asyncio
import sys
import threading
from collections.abc import Callable

from .consumer import MessageConsumer

CAN_USE_ASYNCIO_RUNNER = sys.version_info >= (3, 11)


class MessageConsumerThreadRunner:
    def __init__(self, message_consumer_factory: Callable[[], MessageConsumer]):
        self.message_consumer_factory = message_consumer_factory

    def run(self, num_workers: int = 1) -> None:
        threads = []
        for _ in range(num_workers):
            t = threading.Thread(target=self._run_in_thread)
            t.start()
            threads.append(t)

        for t in threads:
            t.join()

    def _run_in_thread(self) -> None:
        if CAN_USE_ASYNCIO_RUNNER:
            with asyncio.Runner() as runner:
                runner.run(self._run_consumer())
        else:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            loop.run_until_complete(self._run_consumer())

    async def _run_consumer(self) -> None:
        message_consumer = self.message_consumer_factory()
        await message_consumer.run()



from typing import Any

from .base import Filter, Operator, Stage


class Pipeline(Operator):
    def __init__(self, *stages: Stage):
        self.stages = list(stages)

    def expression(self) -> list[dict[str, Any]]:
        return Filter.non_empty(*self.stages).expression()



from . import (
    command,
    compat,
    dependency,
    entity,
    event,
    messaging,
    pagination,
    query,
    repository,
    result,
    schema,
    types,
    uow,
)
from .container import BaseModel, TimeStampedModel

__all__ = [
    "BaseModel",
    "TimeStampedModel",
    "entity",
    "field",
    "schema",
    "repository",
    "query",
    "types",
    "command",
    "event",
    "dependency",
    "compat",
    "pagination",
    "result",
    "uow",
    "messaging",
]



from __future__ import annotations

import enum
import inspect
from typing import Any, ClassVar, Generic, Optional, TypeVar, get_args, get_type_hints

import pymongo
from pymongo import UpdateOne
from pymongo.client_session import ClientSession
from pymongo.collection import Collection
from pymongo.database import Database

from .. import compat, types
from ..entity import Entity
from ..pagination import Pagination
from ..query.mongo import MongoFilterQuery, MongoQuery
from . import base

EntityType = TypeVar("EntityType", bound=Entity)


class SessionRequirement(str, enum.Enum):
    REQUIRED = "required"
    OPTIONAL = "optional"


def entity_to_mongo_schema(entity: EntityType, **kwargs) -> dict[str, Any]:
    default_kwargs = {"by_alias": True}
    return entity.model_dump(**(default_kwargs | kwargs))


class MongoRepository(base.BaseRepository, Generic[EntityType]):
    """
    `MongoRepository` 추상 클래스

    상속 인자:
        `session_requirement`: `MongoRepository` 상속 시 `session_requirement` 인자를 통해
        `session` 인자 선언 여부를 검사할 수 있습니다. 프로덕션에서는 `session_requirement`를
        `SessionRequirement.REQUIRED`로 설정하는 것을 권장합니다.

        ```python
        class UserRepository(
            MongoRepository[User],
            session_requirement=SessionRequirement.REQUIRED,
        ):
            ...
        ```

    """

    __collection_name__: ClassVar[str] = ""
    __indexes__: ClassVar[Optional[list[pymongo.IndexModel]]] = None

    def __init__(self, collection: Collection):
        self.collection = collection

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        cls._validate_session_requirement(
            kwargs.get("session_requirement", SessionRequirement.OPTIONAL)
        )

    @classmethod
    def _validate_session_requirement(cls, requirement: SessionRequirement) -> None:
        if requirement == SessionRequirement.OPTIONAL:
            return

        error_message = "Method {} does not have a session argument"
        # 클래스 자체를 검사하기 때문에 `inspect.ismethod`를 사용하면 인스턴스 메서드가 아니라 클래스메서드를 리턴
        for name, member in inspect.getmembers(cls, predicate=inspect.isfunction):
            if name.startswith("__"):
                continue

            # https://stackoverflow.com/q/72510518
            # `inspect.get_annotations`를 사용하면 `__future__.annotations`을 사용했을 때에는 문자열을,
            # 그렇지 않은 경우에는 `type`을 리턴함. `__future__.annotations` 사용 여부에 관계 없이
            # 항상 `type`을 얻기 위해 `typing.get_type_hints` 사용
            if (type_hints := get_type_hints(member).get("session")) is None:
                raise ValueError(error_message.format(name))

            if not (args := get_args(type_hints)):
                raise ValueError(error_message.format(name))

            session = args[0]
            if session is not ClientSession:
                raise ValueError(error_message.format(name))

    @classmethod
    def create(cls, database: Database, **kwargs) -> MongoRepository:
        collection = database.get_collection(cls.__collection_name__, **kwargs)
        if cls.__indexes__ is not None:
            collection.create_indexes(cls.__indexes__)
        return cls(collection)

    def find_by_id(
        self,
        entity_id: types.PyObjectId,
        session: ClientSession | None = None,
        **kwargs,
    ) -> Optional[EntityType]:
        return self.find_by({"_id": entity_id}, session=session, **kwargs)

    def find_by(
        self, filter_: dict[str, Any], session: ClientSession | None = None, **kwargs
    ) -> Optional[EntityType]:
        """https://stackoverflow.com/a/73746554/9331155"""
        entity_type: EntityType = get_args(self.__class__.__orig_bases__[0])[0]  # type: ignore
        result = self.collection.find_one(filter=filter_, session=session, **kwargs)
        return result and compat.model_validate(entity_type, result)

    def find_by_query(
        self, qry: MongoQuery, session: ClientSession | None = None, **kwargs
    ) -> Optional[dict[str, Any]]:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        return next(query_result, None)

    def list_by_query(
        self, qry: MongoQuery, session: ClientSession | None = None, **kwargs
    ) -> list[dict[str, Any]]:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        return list(query_result)

    def add(self, entity: EntityType, session: ClientSession | None = None, **kwargs) -> None:
        self.collection.insert_one(entity_to_mongo_schema(entity), session=session, **kwargs)

    def add_many(
        self, entities: list[EntityType], session: ClientSession | None = None, **kwargs
    ) -> None:
        self.collection.insert_many(
            [entity_to_mongo_schema(entity) for entity in entities], session=session, **kwargs
        )

    def update(self, entity: EntityType, session: ClientSession | None = None, **kwargs) -> None:
        self.collection.update_one(
            {"_id": entity.id},
            {"$set": entity_to_mongo_schema(entity)},
            session=session,
            **kwargs,
        )

    def update_many(
        self, entities: list[EntityType], session: ClientSession | None = None, **kwargs
    ) -> None:
        self.collection.bulk_write(
            requests=[
                UpdateOne({"_id": entity.id}, {"$set": entity_to_mongo_schema(entity)})
                for entity in entities
            ],
            session=session,
            **kwargs,
        )

    def delete(
        self, entity_id: types.PyObjectId, session: ClientSession | None = None, **kwargs
    ) -> None:
        self.collection.delete_one({"_id": entity_id}, session=session, **kwargs)

    def execute_raw(
        self,
        operation: str,
        session: ClientSession | None = None,
        **operation_kwargs,
    ) -> Any:
        op = getattr(self.collection, operation, None)
        if op is None:
            raise ValueError(f"Unknown operation on collection: {operation}")
        return op(session=session, **operation_kwargs)

    def filter(
        self, qry: MongoFilterQuery, session: ClientSession | None = None, **kwargs
    ) -> Pagination:
        query_result = self.collection.aggregate(qry.to_query(), session=session, **kwargs)
        if (unwrapped := next(query_result, None)) is None:
            raise ValueError(f"{qry.__class__.__name__} returned nothing")

        return Pagination(
            total=(unwrapped["metadata"][0]["total"] if unwrapped["metadata"] else 0),
            items=unwrapped["items"],
            page=qry.page,
            per_page=qry.per_page,
        )



from .consumer import MessageConsumer
from .consumer_runner import MessageConsumerThreadRunner
from .model import EventMessage, SqsEventMessage
from .queue import MessageQueue, SqsMessageQueue

__all__ = [
    "EventMessage",
    "SqsEventMessage",
    "MessageQueue",
    "SqsMessageQueue",
    "MessageConsumer",
    "MessageConsumerThreadRunner",
]



import asyncio
import sys
from collections.abc import Callable
from typing import Literal

from . import model
from .messagebus import MessageBus
from .queue import MessageQueue

CAN_USE_ASYNCIO_RUNNER = sys.version_info >= (3, 11)

HookEventType = Literal["on_receive", "on_receive_error", "on_consume", "on_consume_error"]
Hook = Callable[[model.EventMessage | Exception], None]


class MessageConsumer:
    def __init__(
        self,
        messagebus: MessageBus,
        message_queue: MessageQueue,
        hooks: dict[HookEventType, list[Hook]] | None = None,
    ):
        self.messagebus = messagebus
        self.message_queue = message_queue
        self.hooks = hooks or {}

        self._default_hook = lambda _: None

    async def run(self) -> None:
        while True:
            try:
                message = self.message_queue.peek()
            except Exception as exc:
                self._execute_hook(hook_event_type="on_receive_error", message=exc)
                continue

            self._execute_hook(hook_event_type="on_receive", message=message)
            if message is None:
                continue

            try:
                await asyncio.create_task(self.consume(message))
            except Exception as exc:
                self._execute_hook(hook_event_type="on_consume_error", message=exc)
                continue
            self._execute_hook(hook_event_type="on_consume", message=message)

    async def consume(self, message: model.EventMessage) -> None:
        await self.messagebus.handle_event(message.body)
        self.message_queue.delete(message)

    def _execute_hook(
        self,
        hook_event_type: HookEventType,
        message: model.EventMessage | Exception,
    ) -> None:
        for hook in self.hooks.get(hook_event_type, [self._default_hook]):
            hook(message)



from __future__ import annotations

import abc
import functools
import operator
from collections.abc import Callable
from typing import Any, ClassVar, Optional

from .types import DictExpression, ListExpression


class Operator:
    @abc.abstractmethod
    def expression(self) -> Any:
        raise NotImplementedError


class ComparisonOperator(Operator):
    def __init__(self, field: str, value: Optional[Any] = None):
        self.field = field
        self.value = value

    @abc.abstractmethod
    def expression(self) -> dict[str, Any]:
        raise NotImplementedError


class LogicalOperator(Operator):
    def __init__(self, *ops: Operator):
        self.ops = list(ops)

    @abc.abstractmethod
    def expression(self) -> dict[str, ListExpression]:
        raise NotImplementedError


class GeneralAggregationOperator(Operator):
    mongo_operator: ClassVar[str] = ""

    def __init__(self, *expressions: Any):
        self.expressions = list(expressions)

    def expression(self) -> DictExpression:
        return {self.mongo_operator: [Evaluable(expr).expression() for expr in self.expressions]}


class Stage(Operator):
    @abc.abstractmethod
    def expression(self) -> DictExpression:
        raise NotImplementedError


class Merge(Operator):
    def __init__(self, *ops: Operator, initial: Any):
        self.ops = list(ops)
        self.initial = initial

    def expression(self) -> Any:
        return functools.reduce(operator.or_, [op.expression() for op in self.ops], self.initial)

    @classmethod
    def dict(cls, *ops: Operator) -> Merge:
        return cls(*ops, initial={})


class Filter(Operator):
    def __init__(self, *ops: Operator, predicate: Callable[[Operator], bool]):
        self.ops = list(ops)
        self.predicate = predicate

    def expression(self) -> ListExpression:
        return [op.expression() for op in self.ops if self.predicate(op)]

    @classmethod
    def non_empty(cls, *ops: Operator) -> Filter:
        return cls(*ops, predicate=lambda op: op.expression())


class Evaluable(Operator):
    def __init__(self, op: Any):
        self.op = op

    def expression(self) -> Any:
        return self.op.expression() if isinstance(self.op, Operator) else self.op



from collections.abc import Generator
from typing import TypeVar

T = TypeVar("T")


def descendants_of(cls: type[T]) -> Generator[T, None, None]:
    stack = cls.__subclasses__()
    while stack:
        current_cls = stack.pop()
        yield current_cls
        stack.extend(current_cls.__subclasses__())



from typing import Any

from ..base import GeneralAggregationOperator, Operator
from ..types import DictExpression


class Concat(GeneralAggregationOperator):
    mongo_operator = "$concat"


class ToString(Operator):
    def __init__(self, expr: Any):
        self.expr = expr

    def expression(self) -> DictExpression:
        return {"$toString": self.expr}



import enum
from typing import Any, ClassVar, Optional, Union


class BaseError(Exception):
    default_message: ClassVar[Optional[str]] = None

    def __init__(
        self,
        message: Optional[Union[str, enum.Enum]] = None,
        detail: Optional[Any] = None,
        invalid_params: Optional[list[dict[str, Any]]] = None,
    ):
        if isinstance(message, enum.Enum):
            message = message.value
        super().__init__(message)

        if message is None and self.default_message is None:
            raise ValueError("`message` or `default_message` must not be None")

        self.message = str(message) if message is not None else self.default_message
        self.detail = detail
        self.invalid_params = invalid_params

    def __str__(self):
        return self.message



from .. import compat
from .datetime import DateTime
from .object_id import PyObjectId

__all__ = [
    "PyObjectId",
    "DateTime",
]

if compat.IS_PYDANTIC_V2:
    from .helper import SupportsGetValidators, chain, get_pydantic_core_schema  # noqa: F401

    __all__.extend(["SupportsGetValidators", "get_pydantic_core_schema", "chain"])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['compose',
 'compose.command',
 'compose.dependency',
 'compose.event',
 'compose.messaging',
 'compose.messaging.queue',
 'compose.query',
 'compose.query.mongo',
 'compose.query.mongo.op',
 'compose.query.mongo.op.aggregation',
 'compose.repository',
 'compose.schema',
 'compose.types',
 'compose.uow']

package_data = \
{'': ['*']}

install_requires = \
['dependency-injector>=4.41.0,<5.0.0',
 'inflection>=0.5.1,<0.6.0',
 'pendulum>=2.1.2,<3.0.0',
 'pydantic>=1.10.0,<3',
 'pymongo[aws]>=4.3.3,<5.0.0']

extras_require = \
{'aws': ['boto3>=1.28.73,<2.0.0']}

setup_kwargs = {
    'name': 'pozalabs-compose',
    'version': '0.21.0',
    'description': 'Backend components for POZAlabs',
    'long_description': 'None',
    'author': 'sunwoong',
    'author_email': 'sunwoong@pozalabs.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from .base import Operator
from .types import DictExpression


class Expr(Operator):
    def __init__(self, op: Operator):
        self.op = op

    def expression(self) -> DictExpression:
        return {"$expr": self.op.expression()}
